# Copyright 2021, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Standard DpEvent classes.

A `DpEvent` represents the (hyper)parameters of a differentially
private query, amplification mechanism, or composition, that are necessary
and sufficient for privacy accounting. Various independent implementations of DP
algorithms that are functionally equivalent from an accounting perspective may
correspond to the same `DpEvent`. Similarly, various independent implementations
of accounting algorithms may consume the same `DpEvent`.

All `DpEvents` processed together are assumed to take place on a single dataset
of records. `DpEvents` fall into roughly three categories:
 - `DpEvents` that release an output, and incur a privacy cost,
    e.g., `GaussianDpEvent`.
 - `DpEvents` that select a subset (or subsets) of the dataset, and run nested
    `DpEvents` on those subsets, e.g., `PoissonSampledDpEvent`.
 - `DpEvents` that represent (possibly sequentially) applying (multiple)
   mechanisms to the dataset (or currently active subset). Currently, this is
   only `ComposedDpEvent` and `SelfComposedDpEvent`.

Each `DpEvent` should completely document the mathematical behavior and
assumptions of the mechanism it represents so that the writer of an accountant
class can implement the accounting correctly without knowing any other
implementation details of the algorithm that produced it.

New mechanism types should be given a corresponding `DpEvent` class, although
not all accountants will be required to support them. In general,
`PrivacyAccountant` implementations are not required to be aware of all
`DpEvent` classes, but they should support the following basic events and handle
them appropriately: `NoOpDpEvent`, `NonPrivateDpEvent`, `ComposedDpEvent`, and
`SelfComposedDpEvent`. They should return `supports(event)` is False for
`UnsupportedDpEvent` or any other event type they have not been designed to
handle.

To ensure that a `PrivacyAccountant` does not accidentally start to return
incorrect results, the following should be enforced:
 * `DpEvent` classes and their parameters should never be removed, barring some
   extended, onerous deprecation process.
 * New parameters cannot be added to existing mechanisms unless they are
   optional. That is, old composed `DpEvent` objects that do not include them
   must remain valid.
 * The meaning of existing mechanisms or parameters must not change. That is,
   existing mechanisms should not have their implementations change in ways that
   alter their privacy properties; new `DpEvent` classes should be added
   instead.
 * `PrivacyAccountant` implementations are expected to return `supports(event)`
   is `False` when processing unknown mechanisms.
"""

from collections.abc import Mapping, Sequence
import importlib
import typing
from typing import List, NamedTuple, Protocol, Union

import attr


@typing.runtime_checkable
class DpEventNamedTuple(Protocol):
  _fields: tuple[str, ...]

  module_name: str
  class_name: str


@attr.s(frozen=True)
class DpEvent(object):
  """Represents application of a private mechanism.

  A `DpEvent` describes a differentially private mechanism sufficiently for
  computing the associated privacy losses, both in isolation and in combination
  with other `DpEvent`s.
  """

  def to_named_tuple(self) -> DpEventNamedTuple:
    """Converts DpEvent to a NamedTuple representation.

    This can be useful for passing around DpEvents in contexts where it is
    undesirable to take dp_accounting as a dependency, or for serialization.

    Returns:
      NamedTuple representing the DpEvent.

    Raises:
      ValueError: This type of DpEvent is not convertable to NamedTuple. This
        could happen for example if the DpEvent has private fields or has a
        field with init = False.
    """
    cls = type(self)
    _check_attrs_cls_for_known_errors(cls)

    fields = [('module_name', str), ('class_name', str)]
    fields.extend([(x.name, x.type) for x in attr.fields(cls)])
    named_tuple_wrapper_name = f'_{cls.__name__}NamedTupleWrapper'
    _NamedTupleWrapper = NamedTuple(named_tuple_wrapper_name, fields)  # pylint: disable=invalid-name

    def _to_named_tuple(value):
      if attr.has(type(value)):
        value = value.to_named_tuple()
      return value

    values = {'module_name': cls.__module__, 'class_name': cls.__name__}
    for key, value in attr.asdict(self, recurse=False).items():
      if attr.has(type(value)):
        value = _to_named_tuple(value)
      elif isinstance(value, Mapping):
        elements = [(k, _to_named_tuple(v)) for k, v in value.items()]
        mapping_type = type(value)
        value = mapping_type(elements)
      elif isinstance(value, Sequence):
        elements = [_to_named_tuple(x) for x in value]
        sequence_type = type(value)
        value = sequence_type(elements)
      values[key] = value

    return _NamedTupleWrapper(**values)

  @classmethod
  def from_named_tuple(cls, obj: DpEventNamedTuple) -> 'DpEvent':
    """Converts NamedTuple representation to corresponding DpEvent.

    This can be useful for passing around DpEvents in contexts where it is
    undesirable to take dp_accounting as a dependency, or for serialization.

    Args:
      obj: The NamedTuple to convert to a DpEvent.

    Returns:
      The DpEvent corresponding to the NamedTuple.

    Raises:
      ValueError: This type of DpEvent is not convertable to NamedTuple. This
        could happen for example if the DpEvent has private fields or has a
        field with init = False.
    """

    def _from_named_tuple(value):
      if isinstance(value, DpEventNamedTuple):
        value = DpEvent.from_named_tuple(value)
      return value

    fields = set(obj._fields) - set(['module_name', 'class_name'])
    values = {}
    for field in fields:
      value = getattr(obj, field)
      if isinstance(value, DpEventNamedTuple):
        value = _from_named_tuple(value)
      elif isinstance(value, Mapping):
        elements = [(k, _from_named_tuple(v)) for k, v in value.items()]
        mapping_type = type(value)
        value = mapping_type(elements)
      elif isinstance(value, Sequence):
        elements = [_from_named_tuple(x) for x in value]
        sequence_type = type(value)
        value = sequence_type(elements)
      values[field] = value

    module_name = obj.module_name
    if isinstance(module_name, bytes):
      module_name = module_name.decode()
    class_name = obj.class_name
    if isinstance(class_name, bytes):
      class_name = class_name.decode()
    module = importlib.import_module(module_name)
    subcls = getattr(module, class_name)

    try:
      return subcls(**values)
    except Exception as e:
      _check_attrs_cls_for_known_errors(subcls)
      raise e


def _check_attrs_cls_for_known_errors(cls: type[DpEvent]) -> None:
  """Checks properties of attrs that are not compatible with NamedTuple.

  Args:
    cls: Class type to check.

  Raises:
    ValueError: cls is incompatible with conversion to NamedTuple.
  """
  if not attr.has(cls):
    raise ValueError(
        f'Expected `cls` to be an `attrs` decorated class, found `{cls}`.'
    )
  for field in attr.fields(cls):
    if field.name.startswith('_'):
      raise ValueError(
          'Expected all fields on the `attrs` decorated class to be public, '
          f'found `{cls}` has a field `{field.name}` thats starts with `_`, '
          'making it private.'
      )
    if not field.init:
      raise ValueError(
          'Expected all fields on the `attrs` decorated class to set `init` to '
          f'True, found `{cls}` has a field `{field.name}` that set `init` to '
          'False.'
      )


@attr.s(frozen=True)
class NoOpDpEvent(DpEvent):
  """Represents appplication of an operation with no privacy impact.

  A `NoOpDpEvent` is generally never required, but it can be useful as a
  placeholder where a `DpEvent` is expected, such as in tests or some live
  accounting pipelines.
  """


@attr.s(frozen=True)
class NonPrivateDpEvent(DpEvent):
  """Represents application of a non-private operation.

  This `DpEvent` should be used when an operation is performed that does not
  satisfy (epsilon, delta)-DP. All `PrivacyAccountant`s should return infinite
  epsilon/delta when encountering a `NonPrivateDpEvent`.
  """


@attr.s(frozen=True)
class UnsupportedDpEvent(DpEvent):
  """Represents application of an as-yet unsupported operation.

  This `DpEvent` should be used when an operation is performed that does not yet
  have any associated DP description, or if the description is temporarily
  inaccessible, for example, during development. All `PrivacyAccountant`s should
  return `supports(event) == False` for `UnsupportedDpEvent`.
  """


@attr.s(frozen=True, slots=True, auto_attribs=True)
class RandomizedResponseDpEvent(DpEvent):
  """Represents application of the randomized response mechanism.

  The Randomized Response over k buckets with noise parameter p takes in an
  input which is one of the k buckets. With probability 1 - p, it simply
  outputs the input bucket. Otherwise, with probability p, it outputs a bucket
  drawn uniformly at random from the k buckets.

  The noise parameter p can be any value in [0, 1], with p=0 corresponding to
  the case where the mechanism always outputs the input bucket, and p=1
  corresponding to the case where the mechanism outputs a bucket drawn
  uniformly at random from the k buckets regardless of the input bucket.
  """
  noise_parameter: float
  num_buckets: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class GaussianDpEvent(DpEvent):
  """Represents an application of the Gaussian mechanism.

  For values v_i and noise z ~ N(0, s^2I), this mechanism returns sum_i v_i + z.
  If the norms of the values are bounded ||v_i|| <= C, the noise_multiplier is
  defined as s / C.
  """
  noise_multiplier: float


@attr.s(frozen=True, slots=True, auto_attribs=True)
class LaplaceDpEvent(DpEvent):
  """Represents an application of the Laplace mechanism.

  For values v_i and noise z sampled coordinate-wise from the Laplace
  distribution L(0, s), this mechanism returns sum_i v_i + z.
  The probability density function of the Laplace distribution L(0, s) with
  parameter s is given as exp(-|x|/s) * (0.5/s) at x for any real value x.
  If the L_1 norm of the values are bounded ||v_i||_1 <= C, the noise_multiplier
  is defined as s / C.
  """
  noise_multiplier: float


@attr.s(frozen=True, slots=True, auto_attribs=True)
class DiscreteLaplaceDpEvent(DpEvent):
  """Represents an application of the discrete Laplace mechanism.

  For integer values v_i and noise z sampled coordinate-wise from the discrete
  Laplace distribution DLap(a), this mechanism returns sum_i v_i + z.
  The probability mass function of the discrete Laplace distribution DLap(a)
  with noise parameter `a` is given as exp(-a|x|) * tanh(a/2) at x for any
  integer value x. If the L_1 norm of the values are bounded ||v_i||_1 <= C,
  the sensitivity is `C`.
  """
  noise_parameter: float
  sensitivity: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class SelfComposedDpEvent(DpEvent):
  """Represents repeated application of a mechanism.

  The repeated applications may be adaptive, where the query producing each
  event depends on the results of prior queries.

  This is equivalent to `ComposedDpEvent` that contains a list of length `count`
  of identical copies of `event`.
  """
  event: DpEvent
  count: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ComposedDpEvent(DpEvent):
  """Represents application of a series of composed mechanisms.

  The composition may be adaptive, where the query producing each event depends
  on the results of prior queries.
  """
  events: List[DpEvent]


@attr.s(frozen=True, slots=True, auto_attribs=True)
class PoissonSampledDpEvent(DpEvent):
  """Represents an application of Poisson subsampling.

  Each record in the dataset is included in the sample independently with
  probability `sampling_probability`. Then the `DpEvent` `event` is applied
  to the sample of records.
  """
  sampling_probability: float
  event: DpEvent


@attr.s(frozen=True, slots=True, auto_attribs=True)
class SampledWithReplacementDpEvent(DpEvent):
  """Represents sampling a fixed sized batch of records with replacement.

  A sample of `sample_size` (possibly repeated) records is drawn uniformly at
  random from the set of possible samples of a source dataset of size
  `source_dataset_size`. Then the `DpEvent` `event` is applied to the sample of
  records.
  """
  source_dataset_size: int
  sample_size: int
  event: DpEvent


@attr.s(frozen=True, slots=True, auto_attribs=True)
class SampledWithoutReplacementDpEvent(DpEvent):
  """Represents sampling a fixed sized batch of records without replacement.

  A sample of `sample_size` unique records is drawn uniformly at random from the
  set of possible samples of a source dataset of size `source_dataset_size`.
  Then the `DpEvent` `event` is applied to the sample of records.
  """
  source_dataset_size: int
  sample_size: int
  event: DpEvent


@attr.s(frozen=True, slots=True, auto_attribs=True)
class SingleEpochTreeAggregationDpEvent(DpEvent):
  """Represents aggregation for a single epoch using one or more trees.

  Multiple tree-aggregation steps can occur, but it is required that each
  record occurs at most once *across all trees*. See appendix D of
  "Practical and Private (Deep) Learning without Sampling or Shuffling"
  https://arxiv.org/abs/2103.00039.

  To represent the common case where the same record can occur in multiple
  trees (but still at most once per tree), wrap this with `SelfComposedDpEvent`
  or `ComposedDpEvent` and use a scalar for `step_counts`.

  Attributes:
    noise_multiplier: The ratio of the noise per node to the sensitivity.
    step_counts: The number of steps in each tree. May be a scalar for a single
      tree.
  """
  noise_multiplier: float
  step_counts: Union[int, List[int]]


@attr.s(frozen=True, slots=True, auto_attribs=True)
class RepeatAndSelectDpEvent(DpEvent):
  """Represents repeatedly running the mechanism and selecting the best output.

  The total number of runs is randomized and drawn from a distribution
  with the given parameters: Poisson (shape=infinity), Geometric (shape=1),
  Logarithmic (shape=0), or Truncated Negative binomial (0<shape<infinity).

  See https://arxiv.org/abs/2110.03620 for details.

  Attributes:
    event: The DpEvent that is being repeated.
    mean: The mean number of repetitions.
    shape: The shape of the distribution of the number of repetitions.
  """
  event: DpEvent
  mean: float
  shape: float


@attr.s(frozen=True, slots=True, auto_attribs=True)
class MixtureOfGaussiansDpEvent(DpEvent):
  """Represents an application of the Mixture of Gaussians mechanism.

  For sigma, sensitivities c_i and probabilities p_i, given D this mechanism
  outputs a sample from N(0, sigma^2) and given D' this mechanism outputs a
  sample from sum_i p_i N(c_i, sigma^2).

  See https://arxiv.org/abs/2310.15526 for details.

  Attributes:
    standard_deviation: The standard deviation of the Gaussian noise.
    sensitivities: The support of the sensitivity random variable. Should be the
      same length as sampling_probs.
    sampling_probs: The probabilities associated with the sensitivities.
  """

  standard_deviation: float
  sensitivities: Sequence[float]
  sampling_probs: Sequence[float]



# Copyright 2021, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PrivacyAccountant abstract base class."""

import abc
import enum
from typing import Any, Optional

import attr

from dp_accounting import dp_event
from dp_accounting import dp_event_builder


class NeighboringRelation(enum.Enum):
  ADD_OR_REMOVE_ONE = 1
  REPLACE_ONE = 2

  # A record is replaced with a special record, such as the "zero record". See
  # https://arxiv.org/pdf/2103.00039.pdf, Definition 1.1.
  REPLACE_SPECIAL = 3


class UnsupportedEventError(Exception):
  """Exception to raise if _compose is called on unsupported event type."""


class PrivacyAccountant(metaclass=abc.ABCMeta):
  """Abstract base class for privacy accountants."""

  def __init__(self, neighboring_relation: NeighboringRelation):
    self._neighboring_relation = neighboring_relation
    self._ledger = dp_event_builder.DpEventBuilder()

  @property
  def neighboring_relation(self) -> NeighboringRelation:
    """The neighboring relation used by the accountant.

    The neighboring relation is expected to remain constant after
    initialization. Subclasses should not override this property or change the
    value of the private attribute.
    """
    return self._neighboring_relation

  def supports(self, event: dp_event.DpEvent) -> bool:
    """Checks whether the `DpEvent` can be processed by this accountant.

    In general this will require recursively checking the structure of the
    `DpEvent`. In particular `ComposedDpEvent` and `SelfComposedDpEvent` should
    be recursively examined.

    Args:
      event: The `DpEvent` to check.

    Returns:
      True iff this accountant supports processing `event`.
    """
    return self._maybe_compose(event, 0, False) is None

  @attr.s(frozen=True, slots=True, auto_attribs=True)
  class CompositionErrorDetails(object):
    """Describes offending subevent and error in case composition fails."""
    invalid_event: Optional[dp_event.DpEvent]
    error_message: Optional[str]

  @abc.abstractmethod
  def _maybe_compose(self, event: dp_event.DpEvent, count: int,
                     do_compose: bool) -> Optional[CompositionErrorDetails]:
    """Traverses `event` and performs composition if `do_compose` is True.

    If `do_compose` is True, updates internal state to account for application
    of a `DpEvent`. Subsequent calls to `get_epsilon` or `get_delta` will return
    values that account for composition of this `DpEvent`.

    If `do_compose` is False, traverses structure of event to check whether
    composition is supported *without updating internal state*. Returns None if
    composition would succeed, otherwise returns a `CompositionErrorDetails`
    with information about why the composition would fail.

    Args:
      event: A `DpEvent` to process.
      count: The number of times to compose the event.
      do_compose: Whether to actually perform the composition.

    Returns:
      None if composition is valid. If composition is not supported, returns
      `CompositionErrorDetails` describing why the composition fails.
    """

  def compose(self, event: dp_event.DpEvent, count: int = 1) -> Any:
    """Updates internal state to account for application of a `DpEvent`.

    Calls to `get_epsilon` or `get_delta` after calling `compose` will return
    values that account for this `DpEvent`.

    Args:
      event: A `DpEvent` to process.
      count: The number of times to compose the event.

    Returns:
      Reference to this privacy accountant. This allows one to write, e.g.,
      `eps = Accountant().compose(event1).compose(event2).get_epsilon(delta)`.

    Raises:
      UnsupportedEventError: `event` is not supported by this
      `PrivacyAccountant`.

    """
    if not isinstance(event, dp_event.DpEvent):
      raise TypeError(f'`event` must be `DpEvent`. Found {type(event)}.')
    composition_error = self._maybe_compose(event, count, False)
    if composition_error:
      raise UnsupportedEventError(
          f'Unsupported event: {event}. Error: '
          f'[{composition_error.error_message}] caused by subevent '
          f'{composition_error.invalid_event}.')
    self._ledger.compose(event, count)
    self._maybe_compose(event, count, True)
    return self

  @property
  def ledger(self) -> dp_event.DpEvent:
    """Returns the (composed) `DpEvent` processed so far by this accountant."""
    return self._ledger.build()

  @abc.abstractmethod
  def get_epsilon(self, target_delta: float) -> float:
    """Gets the current epsilon.

    Args:
      target_delta: The target delta.

    Returns:
      The current epsilon, accounting for all composed `DpEvent`s.
    """

  def get_delta(self, target_epsilon: float) -> float:
    """Gets the current delta.

    An implementer of `PrivacyAccountant` may choose not to override this, in
    which case `NotImplementedError` will be raised.

    Args:
      target_epsilon: The target epsilon.

    Returns:
      The current delta, accounting for all composed `DpEvent`s.
    """
    raise NotImplementedError()



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for testing.
"""

from typing import Optional, Mapping, Union
import unittest  # pylint:disable=unused-import

import numpy as np


def assert_dictionary_contained(testcase: 'unittest.TestCase',
                                dict1: Mapping[Union[int, float], float],
                                dict2: Mapping[Union[int, float], float]):
  """Check whether first dictionary is contained in the second.

  Keys of type float are checked for almost equality. Values are always checked
  for almost equality.

  Keys corresponding to values close to 0 are ignored in this test.

  Args:
    testcase: unittestTestCase object to assert containment of dictionary.
    dict1: first dictionary
    dict2: second dictionary
  """
  for i in dict1.keys():
    if not np.isclose(dict1[i], 0):
      key_found = False
      value_found = False
      for j in dict2.keys():
        if np.isclose(i, j):
          key_found = True
          if np.isclose(dict1[i], dict2[j]):
            value_found = True
            break
      testcase.assertTrue(key_found,
                          msg=f'Key {i} in {dict1} not found in {dict2}')
      testcase.assertTrue(
          value_found,
          msg=f'Value for key {i} in {dict1} not matching that in {dict2}')


def assert_dictionary_almost_equal(testcase: 'unittest.TestCase',
                                   dictionary1: Mapping[Union[int, float],
                                                        float],
                                   dictionary2: Mapping[Union[int, float],
                                                        float]):
  """Check two dictionaries have almost equal values.

  Keys of type float are checked for almost equality. Values are always checked
  for almost equality.

  Keys corresponding to values close to 0 are ignored in this test.

  Args:
    testcase: unittestTestCase object to assert containment of dictionary.
    dictionary1: first dictionary
    dictionary2: second dictionary
  """
  assert_dictionary_contained(testcase, dictionary1, dictionary2)
  assert_dictionary_contained(testcase, dictionary2, dictionary1)


def assert_almost_greater_equal(testcase: 'unittest.TestCase',
                                a: float, b: float, msg: Optional[str] = None):
  """Asserts that first value is greater or almost equal to second value."""
  msg = f'{a} is less than {b}' if msg is None else msg
  testcase.assertTrue(a >= b or np.isclose(a, b), msg=msg)



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for accountant."""

import unittest

from absl.testing import parameterized

from dp_accounting.pld import accountant
from dp_accounting.pld import common


class AccountantTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'basic_composition',
          'sensitivity': 21,
          'epsilon': 3,
          'delta': 0,
          'num_queries': 10,
          'expected_parameter': 70,
      },
      {
          'testcase_name': 'positive_delta',
          'sensitivity': 1,
          'epsilon': 1,
          'delta': 0.0001,
          'num_queries': 20,
          'expected_parameter': 13.6,
      },
      {
          'testcase_name': 'positive_delta_varying_sensitivity',
          'sensitivity': 0.5,
          'epsilon': 1,
          'delta': 0.0001,
          'num_queries': 20,
          'expected_parameter': 6.8,
      },
      {
          'testcase_name': 'large_num_composition',
          'sensitivity': 1,
          'epsilon': 1,
          'delta': 0.0001,
          'num_queries': 500,
          'expected_parameter': 71.0,
      },)
  def test_get_smallest_laplace_noise(self, epsilon, delta, num_queries,
                                      sensitivity, expected_parameter):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    self.assertAlmostEqual(
        expected_parameter,
        accountant.get_smallest_laplace_noise(
            privacy_parameters, num_queries, sensitivity=sensitivity),
        delta=0.1)

  @parameterized.named_parameters(
      {
          'testcase_name': 'basic_composition',
          'sensitivity': 2,
          'epsilon': 3,
          'delta': 0,
          'num_queries': 5,
          'expected_parameter': 0.3,
      },
      {
          'testcase_name': 'positive_delta',
          'sensitivity': 1,
          'epsilon': 1,
          'delta': 0.0001,
          'num_queries': 20,
          'expected_parameter': 0.073,
      },
      {
          'testcase_name': 'positive_delta_varying_sensitivity',
          'sensitivity': 5,
          'epsilon': 1,
          'delta': 0.0001,
          'num_queries': 20,
          'expected_parameter': 0.014,
      },)
  def test_get_smallest_discrete_laplace_noise(self, epsilon, delta,
                                               num_queries, sensitivity,
                                               expected_parameter):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    self.assertAlmostEqual(
        expected_parameter,
        accountant.get_smallest_discrete_laplace_noise(
            privacy_parameters, num_queries, sensitivity=sensitivity),
        delta=1e-3)

  @parameterized.named_parameters(
      {
          'testcase_name': 'base',
          'sensitivity': 1,
          'epsilon': 1,
          'delta': 0.78760074,
          'num_queries': 1,
          'expected_std': 1/3,
      },
      {
          'testcase_name': 'varying_sensitivity_and_num_queries',
          'sensitivity': 6,
          'epsilon': 1,
          'delta': 0.78760074,
          'num_queries': 25,
          'expected_std': 10,
      })
  def test_get_smallest_gaussian_noise(self, epsilon, delta, num_queries,
                                       sensitivity, expected_std):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    self.assertAlmostEqual(
        expected_std,
        accountant.get_smallest_gaussian_noise(
            privacy_parameters, num_queries, sensitivity=sensitivity))

  @parameterized.named_parameters(
      {
          'testcase_name': 'base',
          'epsilon': 1,
          'delta': 1e-5,
          'num_queries': 1,
          'sensitivity': 1,
          'sampling_prob': 0.01,
          'expected_std': 0.67379294682,
      },
      {
          'testcase_name': 'varying_sensitivity_and_num_queries',
          'epsilon': 1,
          'delta': 1e-5,
          'num_queries': 25,
          'sensitivity': 6,
          'sampling_prob': 0.001,
          'expected_std': 3.20956502598,
      })
  def test_get_smallest_subsampled_gaussian_noise(
      self, epsilon, delta, num_queries, sensitivity, sampling_prob,
      expected_std):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    self.assertAlmostEqual(
        expected_std,
        accountant.get_smallest_subsampled_gaussian_noise(
            privacy_parameters,
            num_queries,
            sensitivity=sensitivity,
            sampling_prob=sampling_prob))

  @parameterized.named_parameters(
      {
          'testcase_name': 'basic_composition',
          'epsilon': 1,
          'delta': 0,
          'total_delta': 0,
          'num_queries': 30,
          'expected_total_epsilon': 30,
      },
      {
          'testcase_name': 'advantage_over_basic1',
          'epsilon': 1,
          'delta': 0.001,
          'total_delta': 0.06,
          'num_queries': 30,
          'expected_total_epsilon': 22,
      },
      {
          'testcase_name': 'advantage_over_basic2',
          'epsilon': 1,
          'delta': 0.001,
          'total_delta': 0.1,
          'num_queries': 30,
          'expected_total_epsilon': 20,
      },
      {
          'testcase_name': 'total_delta_too_small',
          'epsilon': 1,
          'delta': 0.2,
          'total_delta': 0.1,
          'num_queries': 1,
          'expected_total_epsilon': None,
      },
      {
          'testcase_name': 'total_delta_too_small2',
          'epsilon': 1,
          'delta': 0.01,
          'total_delta': 0.26,
          'num_queries': 30,
          'expected_total_epsilon': None,
      })
  def test_advanced_composition(self, epsilon, delta, num_queries, total_delta,
                                expected_total_epsilon):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    total_epsilon = accountant.advanced_composition(privacy_parameters,
                                                    num_queries, total_delta)
    if expected_total_epsilon is None:
      self.assertIsNone(total_epsilon)
    else:
      self.assertAlmostEqual(expected_total_epsilon, total_epsilon)

  @parameterized.named_parameters(
      {
          'testcase_name': 'basic_composition',
          'total_epsilon': 30,
          'total_delta': 0,
          'delta': 0,
          'num_queries': 30,
          'expected_epsilon': 1,
      },
      {
          'testcase_name': 'advantage_over_basic',
          'total_epsilon': 22,
          'total_delta': 0.06,
          'delta': 0.001,
          'num_queries': 30,
          'expected_epsilon': 1,
      },
      {
          'testcase_name': 'advantage_over_basic2',
          'total_epsilon': 5,
          'total_delta': 0.01,
          'delta': 0,
          'num_queries': 50,
          'expected_epsilon': 0.25,
      },
      {
          'testcase_name': 'total_delta_too_small',
          'total_epsilon': 1,
          'total_delta': 0.1,
          'delta': 0.2,
          'num_queries': 1,
          'expected_epsilon': None,
      },
      {
          'testcase_name': 'total_delta_too_small2',
          'total_epsilon': 30,
          'total_delta': 0.26,
          'delta': 0.01,
          'num_queries': 30,
          'expected_epsilon': None,
      })
  def test_get_smallest_epsilon_from_advanced_composition(
      self, total_epsilon, total_delta, num_queries, delta, expected_epsilon):
    total_privacy_parameters = common.DifferentialPrivacyParameters(
        total_epsilon, total_delta)
    epsilon = accountant.get_smallest_epsilon_from_advanced_composition(
        total_privacy_parameters, num_queries, delta)
    if expected_epsilon is None:
      self.assertIsNone(epsilon)
    else:
      self.assertAlmostEqual(expected_epsilon, epsilon, places=6)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Privacy accountant that uses Privacy Loss Distributions."""

import math
from typing import Optional

from dp_accounting import dp_event
from dp_accounting import privacy_accountant
from dp_accounting.pld import privacy_loss_distribution

NeighborRel = privacy_accountant.NeighboringRelation
CompositionErrorDetails = (
    privacy_accountant.PrivacyAccountant.CompositionErrorDetails
)
PLD = privacy_loss_distribution


class PLDAccountant(privacy_accountant.PrivacyAccountant):
  """Privacy accountant that uses Privacy Loss Distributions."""

  def __init__(
      self,
      neighboring_relation: NeighborRel = NeighborRel.ADD_OR_REMOVE_ONE,
      value_discretization_interval: float = 1e-4,
  ):
    super().__init__(neighboring_relation)
    self._contains_non_dp_event = False
    self._pld = PLD.identity(
        value_discretization_interval=value_discretization_interval)
    self._value_discretization_interval = value_discretization_interval

  def _maybe_compose(self, event: dp_event.DpEvent, count: int,
                     do_compose: bool) -> Optional[CompositionErrorDetails]:
    if isinstance(event, dp_event.NoOpDpEvent):
      return None
    elif isinstance(event, dp_event.NonPrivateDpEvent):
      if do_compose:
        self._contains_non_dp_event = True
      return None
    elif isinstance(event, dp_event.SelfComposedDpEvent):
      return self._maybe_compose(event.event, event.count * count, do_compose)
    elif isinstance(event, dp_event.ComposedDpEvent):
      for e in event.events:
        result = self._maybe_compose(e, count, do_compose)
        if result is not None:
          return result
      return None
    elif isinstance(event, dp_event.RandomizedResponseDpEvent):
      if self.neighboring_relation not in [NeighborRel.REPLACE_ONE,
                                           NeighborRel.REPLACE_SPECIAL]:
        error_msg = (
            'neighboring_relation must be `REPLACE_ONE` or '
            '`REPLACE_SPECIAL` for `RandomizedResponseDpEvent`. Found '
            f'{self._neighboring_relation}.')
        return CompositionErrorDetails(
            invalid_event=event, error_message=error_msg)
      if do_compose:
        if event.num_buckets == 1:
          # This is a NoOp event, even when noise_parameter is zero.
          pass
        elif event.noise_parameter == 0:
          self._contains_non_dp_event = True
        else:
          rr_pld = PLD.from_randomized_response(
              noise_parameter=event.noise_parameter,
              num_buckets=event.num_buckets,
              value_discretization_interval=self._value_discretization_interval,
              neighboring_relation=self._neighboring_relation,
          )
          self._pld = self._pld.compose(rr_pld)
      return None
    elif isinstance(event, dp_event.GaussianDpEvent):
      if do_compose:
        if event.noise_multiplier == 0:
          self._contains_non_dp_event = True
        else:
          gaussian_pld = PLD.from_gaussian_mechanism(
              standard_deviation=event.noise_multiplier / math.sqrt(count),
              value_discretization_interval=self._value_discretization_interval,
              neighboring_relation=self.neighboring_relation)
          self._pld = self._pld.compose(gaussian_pld)
      return None
    elif isinstance(event, dp_event.LaplaceDpEvent):
      if self.neighboring_relation not in [
          NeighborRel.ADD_OR_REMOVE_ONE, NeighborRel.REPLACE_SPECIAL
      ]:
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'neighboring_relation must be `ADD_OR_REMOVE_ONE` or '
                '`REPLACE_SPECIAL` for `LaplaceDpEvent`. Found '
                f'{self._neighboring_relation}.'
            ),
        )
      if do_compose:
        if event.noise_multiplier == 0:
          self._contains_non_dp_event = True
        else:
          laplace_pld = PLD.from_laplace_mechanism(
              parameter=event.noise_multiplier,
              value_discretization_interval=self._value_discretization_interval
          ).self_compose(count)
          self._pld = self._pld.compose(laplace_pld)
      return None
    elif isinstance(event, dp_event.DiscreteLaplaceDpEvent):
      if self.neighboring_relation not in [
          NeighborRel.ADD_OR_REMOVE_ONE, NeighborRel.REPLACE_SPECIAL
      ]:
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'neighboring_relation must be `ADD_OR_REMOVE_ONE` or '
                '`REPLACE_SPECIAL` for `DiscreteLaplaceDpEvent`. Found '
                f'{self._neighboring_relation}.'
            ),
        )
      if do_compose:
        if event.noise_parameter == 0:
          self._contains_non_dp_event = True
        else:
          discrete_laplace_pld = PLD.from_discrete_laplace_mechanism(
              parameter=event.noise_parameter,
              sensitivity=event.sensitivity,
              value_discretization_interval=self._value_discretization_interval
          ).self_compose(count)
          self._pld = self._pld.compose(discrete_laplace_pld)
      return None
    elif isinstance(event, dp_event.MixtureOfGaussiansDpEvent):
      if self.neighboring_relation not in [
          NeighborRel.ADD_OR_REMOVE_ONE, NeighborRel.REPLACE_SPECIAL
      ]:
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'neighboring_relation must be `ADD_OR_REMOVE_ONE` or '
                '`REPLACE_SPECIAL` for `MixtureOfGaussiansDpEvent`. Found '
                f'{self._neighboring_relation}.'
            ),
        )
      if do_compose:
        if len(event.sensitivities) == 1 and event.sensitivities[0] == 0.0:
          pass
        elif event.standard_deviation == 0:
          self._contains_non_dp_event = True
        else:
          mog_pld = PLD.from_mixture_gaussian_mechanism(
              standard_deviation=event.standard_deviation,
              sensitivities=event.sensitivities,
              sampling_probs=event.sampling_probs,
              value_discretization_interval=self._value_discretization_interval,
          ).self_compose(count)
          self._pld = self._pld.compose(mog_pld)
      return None
    elif isinstance(event, dp_event.PoissonSampledDpEvent):
      if isinstance(event.event, dp_event.GaussianDpEvent):
        if do_compose:
          if event.sampling_probability == 0:
            pass
          elif event.event.noise_multiplier == 0:
            self._contains_non_dp_event = True
          else:
            subsampled_gaussian_pld = PLD.from_gaussian_mechanism(
                standard_deviation=event.event.noise_multiplier,
                value_discretization_interval=self
                ._value_discretization_interval,
                sampling_prob=event.sampling_probability,
                neighboring_relation=self.neighboring_relation,
            ).self_compose(count)
            self._pld = self._pld.compose(subsampled_gaussian_pld)
        return None
      elif isinstance(event.event, dp_event.LaplaceDpEvent):
        if self.neighboring_relation not in [
            NeighborRel.ADD_OR_REMOVE_ONE, NeighborRel.REPLACE_SPECIAL
        ]:
          return CompositionErrorDetails(
              invalid_event=event,
              error_message=(
                  'neighboring_relation must be `ADD_OR_REMOVE_ONE` or '
                  '`REPLACE_SPECIAL` for `PoissonSampledDpEvent` of a '
                  f'`LaplaceDpEvent`. Found {self._neighboring_relation}.'
              ),
          )
        if do_compose:
          if event.sampling_probability == 0:
            pass
          elif event.event.noise_multiplier == 0:
            self._contains_non_dp_event = True
          else:
            subsampled_laplace_pld = PLD.from_laplace_mechanism(
                parameter=event.event.noise_multiplier,
                value_discretization_interval=self
                ._value_discretization_interval,
                sampling_prob=event.sampling_probability).self_compose(count)
            self._pld = self._pld.compose(subsampled_laplace_pld)
        return None
      else:
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'Subevent of `PoissonSampledEvent` must be either '
                f'`GaussianDpEvent` or `LaplaceDpEvent`. Found {event.event}.'
            ),
        )
    else:
      # Unsupported event (including `UnsupportedDpEvent`).
      return CompositionErrorDetails(
          invalid_event=event, error_message='Unsupported event.')

  def get_epsilon(self, target_delta: float) -> float:
    if self._contains_non_dp_event:
      return math.inf
    return self._pld.get_epsilon_for_delta(target_delta)

  def get_delta(self, target_epsilon: float) -> float:
    if self._contains_non_dp_event:
      return 1
    return self._pld.get_delta_for_epsilon(target_epsilon)  # pytype: disable=bad-return-type



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Privacy accountant that uses Renyi differential privacy."""

import math
from typing import Callable, Optional, Sequence, Tuple, Union

from absl import logging
import numpy as np
from scipy import special

from dp_accounting import dp_event
from dp_accounting import privacy_accountant

NeighborRel = privacy_accountant.NeighboringRelation
CompositionErrorDetails = (
    privacy_accountant.PrivacyAccountant.CompositionErrorDetails
)


def _log_add(logx: float, logy: float) -> float:
  """Adds two numbers in the log space."""
  a, b = min(logx, logy), max(logx, logy)
  if a == -np.inf:  # adding 0
    return b
  # Use exp(a) + exp(b) = (exp(a - b) + 1) * exp(b)
  return math.log1p(math.exp(a - b)) + b  # log1p(x) = log(x + 1)


def _log_sub_sign(logx: float, logy: float) -> Tuple[bool, float]:
  """Returns log(exp(logx)-exp(logy)) and its sign."""
  if logx > logy:
    s = True
    mag = logx + math.log1p(-np.exp(logy - logx))
  elif logx < logy:
    s = False
    mag = logy + np.log1p(-np.exp(logx - logy))
  else:
    s = True
    mag = -np.inf

  return s, mag


def _log_comb(n: float, k: float) -> float:
  """Computes log of binomial coefficient."""
  return (
      special.gammaln(n + 1)
      - special.gammaln(k + 1)
      - special.gammaln(n - k + 1)
  )


def _compute_log_a_int(q: float, sigma: float, alpha: int) -> float:
  """Computes log(A_alpha) for integer alpha, 0 < q < 1."""

  # Initialize with 0 in the log space.
  log_a = -np.inf
  log1mq = math.log1p(-q)

  for i in range(alpha + 1):
    log_coef_i = _log_comb(alpha, i) + i * math.log(q) + (alpha - i) * log1mq
    s = log_coef_i + (i * i - i) / (2 * (sigma**2))
    log_a = _log_add(log_a, s)

  return log_a


_MAX_STEPS_LOG_A_FRAC = 1000


def _compute_log_a_frac(q: float, sigma: float, alpha: float) -> float:
  """Computes log(A_alpha) for fractional alpha, 0 < q < 1."""
  # Computation derived in Sec 3.3 of https://arxiv.org/pdf/1908.10530.
  # The two parts of A_alpha, integrals over (-inf,z0] and [z0, +inf), are
  # initialized to 0 in the log space:
  log_a0, log_a1 = -np.inf, -np.inf
  z0 = sigma**2 * math.log(1 / q - 1) + 0.5
  log1mq = math.log1p(-q)

  last_s0 = last_s1 = -np.inf

  for i in range(_MAX_STEPS_LOG_A_FRAC):
    log_coef = _log_comb(alpha, i)
    j = alpha - i

    log_t0 = log_coef + i * math.log(q) + j * log1mq
    log_t1 = log_coef + j * math.log(q) + i * log1mq

    log_e0 = math.log(0.5) + _log_erfc((i - z0) / (math.sqrt(2) * sigma))
    log_e1 = math.log(0.5) + _log_erfc((z0 - j) / (math.sqrt(2) * sigma))

    log_s0 = log_t0 + (i * i - i) / (2 * (sigma**2)) + log_e0
    log_s1 = log_t1 + (j * j - j) / (2 * (sigma**2)) + log_e1

    log_a0 = _log_add(log_a0, log_s0)
    log_a1 = _log_add(log_a1, log_s1)

    total = _log_add(log_a0, log_a1)

    # Terminate when both s0 and s1 are decreasing and sufficiently small
    # relative to total.
    if (
        log_s0 < last_s0
        and log_s1 < last_s1
        and max(log_s0, log_s1) < total - 30
    ):
      return total

    last_s0 = log_s0
    last_s1 = log_s1

  logging.warning(
      '_compute_log_a_frac failed to converge after %d iterations with q=%f'
      ', sigma=%f, alpha=%f. Excluding this order from the epsilon '
      'computation.',
      _MAX_STEPS_LOG_A_FRAC,
      q,
      sigma,
      alpha,
  )
  return np.inf


def _log_erfc(x: float) -> float:
  """Computes log(erfc(x)) with high accuracy for large x."""
  try:
    return math.log(2) + special.log_ndtr(-x * 2**0.5)
  except NameError:
    # If log_ndtr is not available, approximate as follows:
    r = special.erfc(x)
    if r == 0.0:
      # Using the Laurent series at infinity for the tail of the erfc function:
      #     erfc(x) ~ exp(-x^2-.5/x^2+.625/x^4)/(x*pi^.5)
      # To verify in Mathematica:
      #     Series[Log[Erfc[x]] + Log[x] + Log[Pi]/2 + x^2, {x, Infinity, 6}]
      return (
          -math.log(math.pi) / 2
          - math.log(x)
          - x**2
          - 0.5 * x**-2
          + 0.625 * x**-4
          - 37.0 / 24.0 * x**-6
          + 353.0 / 64.0 * x**-8
      )
    else:
      return math.log(r)


def compute_delta(
    orders: Sequence[float], rdp: Sequence[float], epsilon: float
) -> Tuple[float, float]:
  """Computes delta given a list of RDP values and target epsilon.

  Args:
    orders: An array of orders.
    rdp: An array of RDP guarantees.
    epsilon: The target epsilon.

  Returns:
    2-tuple containing optimal delta and the optimal order.

  Raises:
    ValueError: If input is malformed.
  """
  if epsilon < 0:
    raise ValueError(f'Epsilon cannot be negative. Found {epsilon}.')
  if len(orders) != len(rdp):
    raise ValueError('Input lists must have the same length.')

  # Basic bound (see https://arxiv.org/abs/1702.07476 Proposition 3 in v3):
  #   delta = min( np.exp((rdp - epsilon) * (orders - 1)) )

  # Improved bound from https://arxiv.org/abs/2004.00010 Proposition 12 (in v4):
  logdeltas = []  # work in log space to avoid overflows
  for a, r in zip(orders, rdp):
    if a < 1:
      raise ValueError(f'Renyi divergence order must be at least 1. Found {a}.')
    if r < 0:
      logging.warning(
          'Negative Renyi divergence of %s, probably caused by numerical '
          'instability from extreme DpEvents. Returning a delta of zero.',
          r,
      )
      logdelta = -np.inf
    # For small alpha, we are better of with bound via KL divergence:
    # delta <= sqrt(1-exp(-KL)).
    # Take a min of the two bounds.
    elif r == 0:
      logdelta = -np.inf
    else:
      logdelta = 0.5 * math.log1p(-math.exp(-r))
    if a > 1.01:
      # This bound is not numerically stable as alpha->1.
      # Thus we have a min value for alpha.
      # The bound is also not useful for small alpha, so doesn't matter.
      rdp_bound = (a - 1) * (r - epsilon + math.log1p(-1 / a)) - math.log(a)
      logdelta = min(logdelta, rdp_bound)

    logdeltas.append(logdelta)

  optimal_index = np.argmin(logdeltas)
  return min(math.exp(logdeltas[optimal_index]), 1.0), orders[optimal_index]


def compute_epsilon(
    orders: Sequence[float], rdp: Sequence[float], delta: float
) -> Tuple[float, float]:
  """Computes epsilon given a list of RDP values and target delta.

  Args:
    orders: An array of orders.
    rdp: An array of RDP guarantees.
    delta: The target delta. Must be >= 0.

  Returns:
    2-tuple containing optimal epsilon and the optimal order.

  Raises:
    ValueError: If input is malformed.
  """
  if delta < 0:
    raise ValueError(f'Delta cannot be negative. Found {delta}.')

  if delta == 0:
    if all(r == 0 for r in rdp):
      return 0, 0
    else:
      return np.inf, 0

  if len(orders) != len(rdp):
    raise ValueError('Input lists must have the same length.')

  # Basic bound (see https://arxiv.org/abs/1702.07476 Proposition 3 in v3):
  #   epsilon = min( rdp - math.log(delta) / (orders - 1) )

  # Improved bound from https://arxiv.org/abs/2004.00010 Proposition 12 (in v4).
  # Also appears in https://arxiv.org/abs/2001.05990 Equation 20 (in v1).
  eps = []
  for a, r in zip(orders, rdp):
    if a < 1:
      raise ValueError(f'Renyi divergence order must be at least 1. Found {a}.')
    if r < 0:
      logging.warning(
          'Negative Renyi divergence of %s, probably caused by numerical '
          'instability from extreme DpEvents. Returning an epsilon of zero.',
          r,
      )
      epsilon = 0.0
    elif delta**2 + math.expm1(-r) > 0:
      # In this case, we can simply bound via KL divergence:
      # delta <= sqrt(1-exp(-KL)).
      epsilon = 0  # No need to try further computation if we have epsilon = 0.
    elif a > 1.01:
      # This bound is not numerically stable as alpha->1.
      # Thus we have a min value of alpha.
      # The bound is also not useful for small alpha, so doesn't matter.
      epsilon = r + math.log1p(-1 / a) - math.log(delta * a) / (a - 1)
    else:
      # In this case we can't do anything. E.g., asking for delta = 0.
      epsilon = np.inf
    eps.append(epsilon)

  optimal_index = np.argmin(eps)
  return max(0, eps[optimal_index]), orders[optimal_index]


def _stable_inplace_diff_in_log(
    vec: np.ndarray, signs: np.ndarray, n: Optional[int] = None
):
  """Replaces the first n-1 dims of vec with the log of abs difference operator.

  Args:
    vec: numpy array of floats with size larger than 'n'
    signs: Optional numpy array of bools with the same size as vec in case one
      needs to compute partial differences vec and signs jointly describe a
      vector of real numbers' sign and abs in log scale.
    n: Optonal upper bound on number of differences to compute. If None, all
      differences are computed.

  Returns:
    The first n-1 dimension of vec and signs will store the log-abs and sign of
    the difference.

  Raises:
    ValueError: If input is malformed.
  """

  if vec.shape != signs.shape:
    raise ValueError('Shape of vec and signs do not match.')
  if signs.dtype != bool:
    raise ValueError('signs must be of type bool')
  if n is None:
    n = np.max(vec.shape) - 1
  else:
    assert np.max(vec.shape) >= n + 1
  for j in range(0, n, 1):
    if signs[j] == signs[j + 1]:  # When the signs are the same
      # if the signs are both positive, then we can just use the standard one
      signs[j], vec[j] = _log_sub_sign(vec[j + 1], vec[j])
      # otherwise, we do that but toggle the sign
      if not signs[j + 1]:
        signs[j] = ~signs[j]
    else:  # When the signs are different.
      vec[j] = _log_add(vec[j], vec[j + 1])
      signs[j] = signs[j + 1]


def _get_forward_diffs(
    fun: Callable[[float], float], n: int
) -> Tuple[np.ndarray, np.ndarray]:
  """Computes up to nth order forward difference evaluated at 0.

  See Theorem 27 of https://arxiv.org/pdf/1808.00087.pdf

  Args:
    fun: Function to compute forward differences of.
    n: Number of differences to compute.

  Returns:
    Pair (deltas, signs_deltas) of the log deltas and their signs.
  """
  func_vec = np.zeros(n + 3)
  signs_func_vec = np.ones(n + 3, dtype=bool)

  # ith coordinate of deltas stores log(abs(ith order discrete derivative))
  deltas = np.zeros(n + 2)
  signs_deltas = np.zeros(n + 2, dtype=bool)
  for i in range(1, n + 3, 1):
    func_vec[i] = fun(1.0 * (i - 1))
  for i in range(0, n + 2, 1):
    # Diff in log scale
    _stable_inplace_diff_in_log(func_vec, signs_func_vec, n=n + 2 - i)
    deltas[i] = func_vec[0]
    signs_deltas[i] = signs_func_vec[0]
  return deltas, signs_deltas


def _compute_log_a(
    q: float, noise_multiplier: float, alpha: Union[int, float]
) -> float:
  if float(alpha).is_integer():
    return _compute_log_a_int(q, noise_multiplier, int(alpha))
  else:
    return _compute_log_a_frac(q, noise_multiplier, alpha)


def _compute_rdp_poisson_subsampled_gaussian(
    q: float, noise_multiplier: float, orders: Sequence[float]
) -> Union[float, np.ndarray]:
  """Computes RDP of the Poisson sampled Gaussian mechanism.

  Args:
    q: The sampling rate.
    noise_multiplier: The ratio of the standard deviation of the Gaussian noise
      to the l2-sensitivity of the function to which it is added.
    orders: An array of RDP orders.

  Returns:
    The RDPs at all orders. Can be `np.inf`.

  Raises:
    ValueError: If q is not in [0, 1] or noise_multiplier is negative.
  """
  if not 0 <= q <= 1:
    raise ValueError(f'Sampling rate must be in [0, 1]. Found {q}.')

  if noise_multiplier < 0:
    raise ValueError(
        f'Noise multiplier must be non-negative: {noise_multiplier}'
    )

  def compute_one_order(q, alpha):
    if q == 0:
      return 0

    if np.isinf(alpha) or noise_multiplier == 0:
      return np.inf

    if q == 1.0:
      return alpha / (2 * noise_multiplier**2)

    return _compute_log_a(q, noise_multiplier, alpha) / (alpha - 1)

  return np.array([compute_one_order(q, order) for order in orders])


def _compute_rdp_sample_wor_gaussian(
    q: float, noise_multiplier: float, orders: Sequence[float]
) -> Union[float, np.ndarray]:
  """Computes RDP of Gaussian mechanism using sampling without replacement.

  This function applies to the following schemes:
  1. Sampling w/o replacement: Sample a uniformly random subset of size m = q*n.
  2. ``Replace one data point'' version of differential privacy, i.e., n is
     considered public information.

  Reference: Theorem 27 of https://arxiv.org/pdf/1808.00087.pdf (A strengthened
  version applies subsampled-Gaussian mechanism.)
  - Wang, Balle, Kasiviswanathan. "Subsampled Renyi Differential Privacy and
  Analytical Moments Accountant." AISTATS'2019.

  Args:
    q: The sampling proportion =  m / n.  Assume m is an integer <= n.
    noise_multiplier: The ratio of the standard deviation of the Gaussian noise
      to the l2-sensitivity of the function to which it is added.
    orders: An array of RDP orders.

  Returns:
    The RDPs at all orders, can be np.inf.
  """
  return np.array([
      _compute_rdp_sample_wor_gaussian_scalar(q, noise_multiplier, order)
      for order in orders
  ])


def _compute_rdp_sample_wor_gaussian_scalar(
    q: float, sigma: float, alpha: Union[float, int]
) -> float:
  """Computes RDP of the Sampled Gaussian mechanism at order alpha.

  Args:
    q: The sampling proportion =  m / n.  Assume m is an integer <= n.
    sigma: The std of the additive Gaussian noise.
    alpha: The order at which RDP is computed.

  Returns:
    RDP at alpha, can be np.inf.
  """

  assert (q <= 1) and (q >= 0) and (alpha >= 1)

  if q == 0:
    return 0

  if q == 1.0:
    return alpha / (2 * sigma**2)

  if np.isinf(alpha):
    return np.inf

  if float(alpha).is_integer():
    return _compute_rdp_sample_wor_gaussian_int(q, sigma, int(alpha)) / (
        alpha - 1
    )
  else:
    # When alpha not an integer, we apply Corollary 10 of [WBK19] to interpolate
    # the CGF and obtain an upper bound
    alpha_f = math.floor(alpha)
    alpha_c = math.ceil(alpha)

    x = _compute_rdp_sample_wor_gaussian_int(q, sigma, alpha_f)
    y = _compute_rdp_sample_wor_gaussian_int(q, sigma, alpha_c)
    t = alpha - alpha_f
    return ((1 - t) * x + t * y) / (alpha - 1)


def _compute_rdp_sample_wor_gaussian_int(
    q: float, sigma: float, alpha: int
) -> float:
  """Computes log(A_alpha) for integer alpha, subsampling without replacement.

  When alpha is smaller than max_alpha, compute the bound Theorem 27 exactly,
    otherwise compute the bound with Stirling approximation.

  Args:
    q: The sampling proportion = m / n.  Assume m is an integer <= n.
    sigma: The std of the additive Gaussian noise.
    alpha: The order at which RDP is computed.

  Returns:
    RDP at alpha, can be np.inf.
  """

  max_alpha = 256

  if np.isinf(alpha):
    return np.inf
  elif alpha == 1:
    return 0

  def cgf(x):
    # Return rdp(x+1)*x, the rdp of Gaussian mechanism is alpha/(2*sigma**2)
    return x * 1.0 * (x + 1) / (2.0 * sigma**2)

  def func(x):
    # Return the rdp of Gaussian mechanism
    return 1.0 * x / (2.0 * sigma**2)

  # Initialize with 1 in the log space.
  log_a = 0
  # Calculates the log term when alpha = 2
  log_f2m1 = func(2.0) + math.log1p(-np.exp(-func(2.0)))
  if alpha <= max_alpha:
    # We need forward differences of exp(cgf)
    # The following line is the numerically stable way of implementing it.
    # The output is in polar form with logarithmic magnitude
    deltas, _ = _get_forward_diffs(cgf, alpha)
    # Compute the bound exactly requires book keeping of O(alpha**2)
    for i in range(2, alpha + 1):
      if i == 2:
        s = (
            2 * np.log(q)
            + _log_comb(alpha, 2)
            + np.minimum(np.log(4) + log_f2m1, func(2.0) + np.log(2))
        )
      else:  # i > 2
        delta_lo = deltas[int(2 * np.floor(i / 2.0)) - 1]
        delta_hi = deltas[int(2 * np.ceil(i / 2.0)) - 1]
        s = np.log(4) + 0.5 * (delta_lo + delta_hi)
        s = np.minimum(s, np.log(2) + cgf(i - 1))
        s += i * np.log(q) + _log_comb(alpha, i)
      log_a = _log_add(log_a, s)
    return float(log_a)
  else:
    # Compute the bound with stirling approximation. Everything is O(x) now.
    for i in range(2, alpha + 1):
      if i == 2:
        s = (
            2 * np.log(q)
            + _log_comb(alpha, 2)
            + np.minimum(np.log(4) + log_f2m1, func(2.0) + np.log(2))
        )
      else:
        s = np.log(2) + cgf(i - 1) + i * np.log(q) + _log_comb(alpha, i)
      log_a = _log_add(log_a, s)

    return log_a


def _effective_gaussian_noise_multiplier(
    event: dp_event.DpEvent,
) -> Union[float, dp_event.DpEvent]:
  """Determines the effective noise multiplier of nested structure of Gaussians.

  A series of Gaussian queries on the same data can be reexpressed as a single
  query with pre- and post- processing. For details, see section 3 of
  https://arxiv.org/pdf/1812.06210.pdf.

  Args:
    event: A `dp_event.DpEvent`. In order for conversion to be successful it
      must consist of a single `dp_event.GaussianDpEvent`, or a nested structure
      of `dp_event.ComposedDpEvent` and/or `dp_event.SelfComposedDpEvent`
      bottoming out in `dp_event.GaussianDpEvent`s.

  Returns:
    The noise multiplier of the equivalent `dp_event.GaussianDpEvent`. If the
    input event was not a `dp_event.GaussianDpEvent` or a nested structure of
    `dp_event.ComposedDpEvent` and/or `dp_event.SelfComposedDpEvent` bottoming
    out in `dp_event.GaussianDpEvent`s, returns offending subevent.
  """
  if isinstance(event, dp_event.GaussianDpEvent):
    return event.noise_multiplier
  elif isinstance(event, dp_event.ComposedDpEvent):
    sum_sigma_inv_sq = 0
    for e in event.events:
      sigma = _effective_gaussian_noise_multiplier(e)
      if not isinstance(sigma, float):
        return sigma
      sum_sigma_inv_sq += sigma**-2
    return sum_sigma_inv_sq**-0.5
  elif isinstance(event, dp_event.SelfComposedDpEvent):
    sigma = _effective_gaussian_noise_multiplier(event.event)
    if not isinstance(sigma, float):
      return sigma
    return sigma * event.count**-0.5
  else:
    return event


def _compute_rdp_single_epoch_tree_aggregation(
    noise_multiplier: float,
    step_counts: Union[int, Sequence[int]],
    orders: Sequence[float],
) -> Union[float, np.ndarray]:
  """Computes RDP of the Tree Aggregation Protocol for Gaussian Mechanism.

  This function implements the accounting when the tree is periodically
  restarted and no record occurs twice across all trees. See appendix D of
  "Practical and Private (Deep) Learning without Sampling or Shuffling"
  https://arxiv.org/abs/2103.00039.

  Args:
    noise_multiplier: A non-negative float representing the ratio of the
      standard deviation of the Gaussian noise to the l2-sensitivity of the
      function to which it is added.
    step_counts: A scalar or a list of non-negative integers representing the
      number of steps per epoch (between two restarts).
    orders: An array of RDP orders.

  Returns:
    The RDPs at all orders. Can be `np.inf`.
  """
  if noise_multiplier < 0:
    raise ValueError(
        f'noise_multiplier must be non-negative. Got {noise_multiplier}.'
    )
  if noise_multiplier == 0:
    return np.inf

  if not step_counts:
    raise ValueError(
        'steps_list must be a non-empty list, or a non-zero scalar. Got '
        f'{step_counts}.'
    )

  if np.isscalar(step_counts):
    step_counts = [step_counts]

  for steps in step_counts:
    if steps < 0:
      raise ValueError(f'Steps must be non-negative. Got {step_counts}')

  max_depth = math.ceil(math.log2(max(step_counts) + 1))
  return np.array([a * max_depth / (2 * noise_multiplier**2) for a in orders])


def _expm1_over_x(x: float) -> float:
  """Computes (exp(x)-1)/x in a numerically stable manner.

  Args:
    x: float

  Returns:
    (exp(x)-1)/x
  """
  if x < -0.1 or x > 0.1:
    return math.expm1(x) / x
  # exp(x) = sum_{k>=0} x^k / k!
  # (exp(x)-1)/x = sum_{k>=1} x^{k-1} / k!
  terms = []
  y = 1  # = x^{k-1}/k!
  for k in range(1, 100):
    y = y / k
    terms.append(y)
    y = y * x
  return math.fsum(terms)
  # Dropped terms: sum_{k>=100} x^{k-1} / k!
  # Since |x|<= 0.1, this sum is < 10^-100.
  # Note that 0.9 < (exp(x)-1)/x < 1.1, so this is also a small relative error.


def _logx_over_xm1(x: float) -> float:
  """Computes log(x)/(x-1) in a numerically stable manner.

  Here log is the natural logarithm.

  Args:
    x: float

  Returns:
    log(x)/(x-1)
  """
  if x < 0.9 or x > 1.1:
    return math.log(x) / (x - 1)
  # Denote y = 1-x. Then we have a Taylor series for the natural logarithm:
  # log(x) = log(1-y) = - sum_{k>=1} y^k / k
  # Thus log(x)/(x-1) = -log(1-y)/y = sum_{k>=1} y^{k-1}/k
  return math.fsum((1 - x) ** (k - 1) / k for k in range(1, 100))
  # Dropped terms: sum_{k>=100} y^{k-1}/k.
  # Since |y|<=0.1, this sum is < 10^-100.
  # Since 0.9 <= log(x)/(x-1) <= 1.1, this is also a small relative error.


def _truncated_negative_binomial_mean(gamma: float, shape: float) -> float:
  """Computes the mean of the truncated negative binomial Distribution.

  See Definition 1 in https://arxiv.org/pdf/2110.03620v2.pdf#page=5

  Args:
    gamma: Halting probability parameter of the distribution. Must be >0 and
      <=1.
    shape: Shape parameter of the Distribution. Must be >=0.

  Returns:
    The mean of the distribution.
  """
  if shape < 0:
    raise ValueError(f'shape must be non-negative. Got {shape}')
  if gamma <= 0 or gamma > 1:
    raise ValueError(f'gamma must be in (0,1]. Got {gamma}')

  if shape == 1:  # Geometric Distribution
    return 1 / gamma
  elif shape == 0:  # Logarithmic distribution
    # answer = (1 - 1 / gamma) / log(gamma)
    #        = 1/(gamma*log(gamma)/(gamma-1))
    return 1 / (gamma * _logx_over_xm1(gamma))
  else:  # Truncated Negative Binomial
    # answer = shape * (1 / gamma - 1) / (1 - gamma**shape)
    #        = 1 / ( (exp(shape*log(gamma))-1)/(shape*log(gamma)) *
    #                                              log(gamma)/(gamma-1) * gamma)
    a = _expm1_over_x(shape * math.log(gamma))
    b = _logx_over_xm1(gamma)
    return 1 / (gamma * a * b)


def _gamma_truncated_negative_binomial(
    shape: float, mean: float, tolerance: float = 1e-9
) -> float:
  """Computes gamma parameter of truncated negative binomial from mean.

  Args:
    shape: shape parameter of truncated negative binomial distribution. Must be
      >= 0.
    mean: the expectation of the Distribution
    tolerance: relative (i.e. multiplicative) accuracy bound for gamma. Default
      tolerance is 10^-9.

  Returns:
    The gamma parameter = success probability of the distribution.
  """
  if shape < 0:
    raise ValueError(f'shape must be non-negative. Got {shape}')
  if mean < 1:
    raise ValueError(f'mean must be at least 1. Got {mean}')

  if shape == 1:
    return 1 / mean  # Geometric distribution
  # Otherwise we invert the _truncated_negative_binomial_mean function.
  gamma_min = 0  # gamma=0 corresponds to mean=infinity.
  gamma_max = min(1, 2 * (shape + 1) / mean)  # gamma=1 corresponds to mean=1.
  # Also max{shape,1/ln(1/gamma)}*(1/gamma-1) <= mean <= 2*(shape+1)/gamma,
  # which implies gamma <= 2*(shape+1)/mean
  while gamma_max > gamma_min * (1 + tolerance):
    gamma = (gamma_min + gamma_max) / 2
    gamma_mean = _truncated_negative_binomial_mean(gamma, shape)
    if gamma_mean < mean:
      gamma_max = gamma
    else:
      gamma_min = gamma
  return gamma_min  # The conservative estimate is returned.


def _compute_rdp_repeat_and_select(
    orders: Sequence[float], rdp: Sequence[float], mean: float, shape: float
) -> Sequence[float]:
  # pyformat: disable
  """Computes RDP of repeating and selecting best run.

  Inputs orders & rdp represent RDP of a single run.
  Output represents RDP of running multiple times and returning the best run;
  outputs of other runs are not returned.

  The total number of runs is randomized and drawn from a distribution
  with the given parameters. Poisson (shape=infinity), Geometric (shape=1),
  Logarithmic (shape=0), or Tuncated Negative binomial (0<shape<infinity).

  See https://arxiv.org/abs/2110.03620 for details.

  Args:
    orders: List of Renyi orders considered. Each order should be >= 1.
    rdp: List of RDPs for a single run of the mechanism.
    mean: The mean of the distribution of the random number of repetitions.
    shape: Shape/type of the distribution. Should be >= 0.
       * shape == 0 is the logarithmic distribution.
       * shape == 1 is the geometric distribution.
       * shape == infinity is the Poisson Distribution
       * shape in (0, infinity) is the truncated negative binomial.

  Returns:
    The RDPs at all orders.
  """
  # pyformat: enable
  if math.isnan(shape) or shape < 0:
    raise ValueError(f'Distribution of repetitions must be >=0. Got {shape}.')
  if math.isnan(mean) or mean < 1:
    raise ValueError(f'Mean of number of repetitions must be >=1. Got {mean}.')
  if len(orders) != len(rdp):
    raise ValueError(
        f'orders and rdp must be same length, got {len(orders)} & {len(rdp)}.'
    )

  orders = np.asarray(orders)
  rdp_out = np.zeros_like(orders, dtype=np.float64)  # This will be the output.
  rdp_out += np.inf  # Initialize to infinity.

  if shape == np.inf:  # Poisson Distribution
    for i in range(len(orders)):
      # orders[i]=lambda and rdp[i]=epsilon in the language of
      # Theorem 6 of https://arxiv.org/pdf/2110.03620v2.pdf#page=7
      if orders[i] <= 1:
        continue  # Our formula is not applicable in this case.
      epshat = math.log1p(1 / (orders[i] - 1))
      deltahat, _ = compute_delta(orders, rdp, epshat)
      rdp_out[i] = rdp[i] + mean * deltahat + math.log(mean) / (orders[i] - 1)
  else:  # Truncated Negative Binomial (includes Logarithmic & Geometric)
    # First we map mean parameter to gamma parameter of TNB.
    gamma = _gamma_truncated_negative_binomial(shape, mean)

    # Next we apply the formula.
    # Theorem 2 of https://arxiv.org/pdf/2110.03620v2.pdf#page=5
    # orders[i] = lambda, rdp[i] = epsilon,
    # orders[j] = lambdahat, rdp[j] = epsilonhat
    # First compute constant term
    c = (1 + shape) * np.min((1 - 1 / orders) * rdp - math.log(gamma) / orders)
    for i in range(len(orders)):
      if orders[i] > 1:  # Otherwise our formula is invalid.
        rdp_out[i] = rdp[i] + math.log(mean) / (orders[i] - 1) + c

    # Finally we apply monotonicity of Renyi DP
    # i.e. if orders[i] < orders[j], then rdp[i] < rdp[j].
    # We can use this to bound rdp for low orders.
    for i in range(len(orders)):
      rdp_out[i] = min(
          rdp_out[j] for j in range(len(orders)) if orders[i] <= orders[j]
      )
  return rdp_out


def _laplace_rdp(eps: float, order: float) -> float:
  """Computes RDP of Laplace noise addition.

  See Proposition 6 of Mironov (2017): https://arxiv.org/abs/1702.07476
  RDP = eps + log[ 1 + (a-1) * [exp( (1-2*a) * eps) - 1] / (2*a-1) ] / (a-1).
  In contrast, the naive bound is RDP <= eps, which is tight as order->infinity.
  For small order & eps, we can do better: In general, RDP <= order * eps^2 / 2.
  The above formula is exactly tight for the Laplace mechanism.

  Args:
    eps: The pure DP guarantee corresponding to the limit order->infinity.
    order: The Renyi divergence order (a.k.a. alpha).

  Returns:
    The RDP of Laplace noise addition
  """
  a = float(order)
  eps = float(eps)
  if not math.isfinite(eps) or eps < 0:
    raise ValueError(f'eps must be > 0. Got {eps}')
  if not math.isfinite(a) or a < 1:
    raise ValueError(f'order must be >= 1. Got {a}')

  if a == 1:  # KL divergence
    # Taking the limit as order->1 gives the following expression.
    return eps + math.expm1(-eps)
  elif a < 1.1:  # 1 < order <= 1.1
    # For numerical stability in this case, we use a Taylor series.
    # log(1+x) = sum_{k>=1} -(-x)^k/k
    # Thus log(1+c*x)/x = c * sum_{k>=1} (-c*x)^{k-1}/k
    # Since 0 <= -c*x <= 0.1, this Taylor series converges rapidly.
    c = math.expm1(eps * (1 - 2 * a)) / (2 * a - 1)
    v = -c * (a - 1)
    return eps + c * math.fsum(v ** (k - 1) / k for k in range(1, 100))
  else:  # order > 1.1
    return eps + math.log1p(
        (a - 1) * math.expm1(eps * (1 - 2 * a)) / (2 * a - 1)
    ) / (a - 1)


def _randomized_response_rdp_replace_special(
    p: float, k: int, alpha: float
) -> float:
  """Computes RDP of Randomized Response with REPLACE_SPECIAL relation.

  RDP = max{1/(a-1) * log((k * (1-p) + p)^a / k     + (k - 1) * p^a / k),
            1/(a-1) * log((k * (1-p) + p)^{1-a} / k + (k - 1) * p^{1-a} / k)}.

  Args:
    p: The noise parameter as defined in dp_event.RandomizedResponseDpEvent.
    k: The number of buckets for k-ary Randomized Response mechanism.
    alpha: The Renyi divergence order(s) (a.k.a. alpha).

  Returns:
    The RDP of Randomized Response at order alpha.
  """
  if alpha <= 1:
    raise ValueError(f'Renyi divergence order alpha must be > 1. Got {alpha}')
  if k == 1 or p == 1:
    return 0
  if p == 0:
    return np.inf
  log_1 = np.log(k * (1 - p) + p)
  log_2 = np.log(p)
  if np.isinf(alpha):
    # RDP infinite bound approaches epsilon-DP bound.
    return max(log_1, -log_2)

  bound1, bound2 = (
      special.logsumexp([x * log_1, x * log_2], b=[1 / k, 1 - 1 / k])
      for x in [alpha, 1 - alpha]
  )
  return max(bound1, bound2) / (alpha - 1)


def _randomized_response_rdp_replace_one(
    p: float, k: int, alpha: float
) -> float:
  """Computes RDP of Randomized Response with REPLACE_ONE relation.

  The RDP is the Renyi divergence of order alpha between the following
  distributions supported over {1, 2, ..., k}:
    P = [ 1 - p + p/k, p/k, p/k, ... , p/k ]
    Q = [ p/k, 1 - p + p/k, p/k, ... , p/k ]
  RDP = 1 / (a-1) * log(
      (1 - p + p / k)^a (p / k)^(1 - a) +
      (p / k)^a (1 - p + p/k)^(1 - a) +
      (k - 2) * p / k
    )

  Args:
    p: The noise parameter as defined in dp_event.RandomizedResponseDpEvent.
    k: The number of buckets for k-ary Randomized Response mechanism.
    alpha: The Renyi divergence order(s) (a.k.a. alpha).

  Returns:
    The RDP of Randomized Response at order alpha.
  """
  if alpha <= 1:
    raise ValueError(f'Renyi divergence order alpha must be > 1. Got {alpha}')
  if k == 1 or p == 1:
    return 0
  if p == 0:
    return np.inf

  log_1 = np.log(k / p - k + 1)
  if np.isinf(alpha):
    return log_1
  return special.logsumexp(
      a=[alpha * log_1, -alpha * log_1, 0],
      b=[p / k, 1 - p + p / k, (1 - 2 / k) * p],
  ) / (alpha - 1)


def _compute_randomized_response_rdp(
    noise_parameter: float,
    num_buckets: int,
    alphas: Sequence[float],
    neighboring_relation: NeighborRel,
) -> Union[float, np.ndarray]:
  """Computes RDP of Randomized Response with REPLACE_SPECIAL or REPLACE_ONE neighboring relation for an array of orders.

  Args:
    noise_parameter: The noise parameter as defined in
      dp_event.RandomizedResponseDpEvent.
    num_buckets: The number of buckets for k-ary Randomized Response mechanism.
    alphas: The array of orders.
    neighboring_relation: The neighboring relation.

  Returns:
    The RDP of Randomized Response at given orders.
  """
  if noise_parameter < 0 or noise_parameter > 1:
    raise ValueError(f'noise_parameter must be in [0,1]. Got {noise_parameter}')
  if num_buckets < 1:
    raise ValueError(f'num_buckets must be >= 1. Got {num_buckets}')
  rr_rdp_fn = (
      _randomized_response_rdp_replace_one
      if neighboring_relation == NeighborRel.REPLACE_ONE
      else _randomized_response_rdp_replace_special
  )
  return np.array([
      rr_rdp_fn(p=noise_parameter, k=num_buckets, alpha=alpha)
      for alpha in alphas
  ])


# Default orders chosen to give good coverage for Gaussian mechanism in
# the privacy regime of interest.
DEFAULT_RDP_ORDERS = (
    [1 + x / 10.0 for x in range(1, 100)]
    + list(range(11, 64))
    + [128, 256, 512, 1024]
)


class RdpAccountant(privacy_accountant.PrivacyAccountant):
  """Privacy accountant that uses Renyi differential privacy."""

  def __init__(
      self,
      orders: Optional[Sequence[float]] = None,
      neighboring_relation: NeighborRel = NeighborRel.ADD_OR_REMOVE_ONE,
  ):
    """Initializes the RDP accountant.

    Args:
      orders: The set of RDP orders.
      neighboring_relation: The neighboring relation.
    """
    super().__init__(neighboring_relation)
    if orders is None:
      orders = DEFAULT_RDP_ORDERS
    self._orders = np.array(orders)
    self._rdp = np.zeros_like(orders, dtype=np.float64)

  def _maybe_compose(
      self, event: dp_event.DpEvent, count: int, do_compose: bool
  ) -> Optional[CompositionErrorDetails]:
    """Traverses `event` and performs composition if `do_compose` is True."""

    if isinstance(event, dp_event.NoOpDpEvent):
      return None
    elif isinstance(event, dp_event.NonPrivateDpEvent):
      if do_compose:
        self._rdp += np.inf
      return None
    elif isinstance(event, dp_event.SelfComposedDpEvent):
      return self._maybe_compose(event.event, event.count * count, do_compose)
    elif isinstance(event, dp_event.ComposedDpEvent):
      for e in event.events:
        result = self._maybe_compose(e, count, do_compose)
        if result is not None:
          return result
      return None
    elif isinstance(event, dp_event.GaussianDpEvent):
      if do_compose:
        self._rdp += count * _compute_rdp_poisson_subsampled_gaussian(
            q=1.0, noise_multiplier=event.noise_multiplier, orders=self._orders
        )
      return None
    elif isinstance(event, dp_event.PoissonSampledDpEvent):
      if self._neighboring_relation not in [
          NeighborRel.ADD_OR_REMOVE_ONE,
          NeighborRel.REPLACE_SPECIAL,
      ]:
        error_msg = (
            'neighboring_relation must be `ADD_OR_REMOVE_ONE` or'
            ' `REPLACE_SPECIAL` for `PoissonSampledDpEvent`. Found'
            f' {self._neighboring_relation}.'
        )
        return CompositionErrorDetails(
            invalid_event=event, error_message=error_msg
        )
      sigma_or_bad_event = _effective_gaussian_noise_multiplier(event.event)
      if isinstance(sigma_or_bad_event, dp_event.DpEvent):
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'Subevent of `PoissonSampledDpEvent` must be a'
                ' `GaussianDpEvent` or a nested structure of `ComposedDpEvent`'
                ' and/or `SelfComposedDpEvent` bottoming out in'
                f' `GaussianDpEvent`s. Found subevent {sigma_or_bad_event}.'
            ),
        )
      if do_compose:
        self._rdp += count * _compute_rdp_poisson_subsampled_gaussian(
            q=event.sampling_probability,
            noise_multiplier=sigma_or_bad_event,
            orders=self._orders,
        )
      return None
    elif isinstance(event, dp_event.SampledWithoutReplacementDpEvent):
      if self._neighboring_relation is not NeighborRel.REPLACE_ONE:
        error_msg = (
            'neighboring_relation must be `REPLACE_ONE` for '
            '`SampledWithoutReplacementDpEvent`. Found '
            f'{self._neighboring_relation}.'
        )
        return CompositionErrorDetails(
            invalid_event=event, error_message=error_msg
        )
      sigma_or_bad_event = _effective_gaussian_noise_multiplier(event.event)
      if isinstance(sigma_or_bad_event, dp_event.DpEvent):
        return CompositionErrorDetails(
            invalid_event=event,
            error_message=(
                'Subevent of `SampledWithoutReplacementDpEvent` must be a'
                ' `GaussianDpEvent` or a nested structure of `ComposedDpEvent`'
                ' and/or `SelfComposedDpEvent` bottoming out in'
                f' `GaussianDpEvent`s. Found subevent {sigma_or_bad_event}.'
            ),
        )
      if do_compose:
        self._rdp += count * _compute_rdp_sample_wor_gaussian(
            q=event.sample_size / event.source_dataset_size,
            noise_multiplier=sigma_or_bad_event,
            orders=self._orders,
        )
      return None
    elif isinstance(event, dp_event.SingleEpochTreeAggregationDpEvent):
      if self._neighboring_relation is not NeighborRel.REPLACE_SPECIAL:
        error_msg = (
            'neighboring_relation must be `REPLACE_SPECIAL` for '
            '`SingleEpochTreeAggregationDpEvent`. Found '
            f'{self._neighboring_relation}.'
        )
        return CompositionErrorDetails(
            invalid_event=event, error_message=error_msg
        )
      if do_compose:
        self._rdp += count * _compute_rdp_single_epoch_tree_aggregation(
            event.noise_multiplier, event.step_counts, self._orders
        )
      return None
    elif isinstance(event, dp_event.LaplaceDpEvent):
      if do_compose:
        eps = 1 / event.noise_multiplier
        self._rdp += count * np.array(
            [_laplace_rdp(eps, order) for order in self._orders]
        )
      return None
    elif isinstance(event, dp_event.RepeatAndSelectDpEvent):
      save_rdp = self._rdp
      if do_compose:
        # Save the RDP values from already composed DPEvents. These will
        # be added back after we process this RepeatAndSelectDpEvent.
        # Zero out self._rdp before computing the RDP of the underlying
        # DP event.
        self._rdp = np.zeros_like(self._orders, dtype=np.float64)
      composition_error = self._maybe_compose(event.event, 1, do_compose)
      if composition_error is None and do_compose:
        self._rdp = (
            count
            * _compute_rdp_repeat_and_select(
                self._orders, self._rdp, event.mean, event.shape
            )
            + save_rdp
        )
      return composition_error
    elif isinstance(event, dp_event.RandomizedResponseDpEvent):
      if self._neighboring_relation not in [
          NeighborRel.REPLACE_SPECIAL,
          NeighborRel.REPLACE_ONE,
      ]:
        error_msg = (
            'neighboring_relation must be `REPLACE_SPECIAL` or `REPLACE_ONE`'
            ' for `RandomizedResponseDpEvent`. Found'
            f' {self._neighboring_relation}.'
        )
        return CompositionErrorDetails(
            invalid_event=event, error_message=error_msg
        )
      if do_compose:
        self._rdp += count * _compute_randomized_response_rdp(
            event.noise_parameter,
            event.num_buckets,
            self._orders,
            self._neighboring_relation,
        )
      return None
    else:
      # Unsupported event (including `UnsupportedDpEvent`).
      return CompositionErrorDetails(
          invalid_event=event, error_message='Unsupported event.'
      )

  def get_epsilon_and_optimal_order(
      self, target_delta: float
  ) -> Tuple[float, float]:
    """Returns the current epsilon and the optimal RDP order.

    Args:
      target_delta: The target delta.

    Returns:
      A tuple containing the current epsilon, accounting for all composed
      `DpEvent`s, and the optimal order.
    """
    return compute_epsilon(self._orders, self._rdp, target_delta)

  def get_epsilon(self, target_delta: float) -> float:
    """Returns the current epsilon.

    Args:
      target_delta: The target delta.

    Returns:
      The current epsilon, accounting for all composed `DpEvent`s.
    """
    return compute_epsilon(self._orders, self._rdp, target_delta)[0]

  def get_delta_and_optimal_order(
      self, target_epsilon: float
  ) -> Tuple[float, float]:
    """Returns the current delta and the optimal RDP order.

    Args:
      target_epsilon: The target epsilon.

    Returns:
      A tuple containing the current delta, accounting for all composed
      `DpEvent`s, and the optimal order.
    """
    return compute_delta(self._orders, self._rdp, target_epsilon)

  def get_delta(self, target_epsilon: float) -> float:
    """Returns the current delta.

    Args:
      target_epsilon: The target epsilon.

    Returns:
      The current delta, accounting for all composed `DpEvent`s.
    """
    return compute_delta(self._orders, self._rdp, target_epsilon)[0]

  @property
  def rdp(self) -> np.ndarray:
    """Returns the current RDP values at each order."""
    return np.copy(self._rdp)

  @property
  def orders(self) -> np.ndarray:
    """Returns the set of RDP orders."""
    return np.copy(self._orders)
